import java.util.HashSet;
import java.util.Set;

class Solution {

    static private class SolutionThread extends Thread {
        public SolutionThread(int partial) {
            pos = partial;
            solution = new Solution(partial);
            setName("T" + pos);
        }

        public void run() {
            solution.arrange();
        }

        final Solution solution;
        final int pos;
    }

    public static int solve(final int size, boolean checkTransitions) {

        setLimits(size, false, checkTransitions);
        known_positions.clear();
        solution_pointer = 0;

        Thread solutionThreads[] = new Thread[size];
        for (int i = 0; i < size; i++) {
            solutionThreads[i] = new SolutionThread(i);
            solutionThreads[i].start();
        }

        try {
            for (int i = 0; i < size; i++) {
                solutionThreads[i].join();
            }

        } catch (InterruptedException e) {
        }

        return solution_pointer;
    }

    public static int solveSingle(final int size, boolean first,
            boolean checkTransitions) {
        setLimits(size, first, checkTransitions);
        known_positions.clear();
        solution_pointer = 0;
        (new Solution()).arrange();
        return solution_pointer;

    }

    static private void setLimits(int size, boolean first, boolean check) {
        SIZE = size;
        BITS_PER_CELL = EightQueens.log2(SIZE);
        FILL_MASK = EightQueens.power2(BITS_PER_CELL) - 1;

        EightQueens.log("SIZE=" + SIZE);
        EightQueens.log("BITS_PER_CELL=" + BITS_PER_CELL);
        EightQueens.log("FILL_MASK=" + FILL_MASK);

        FIRST_ONLY = first;
        CHECK_TRANSITIONS = check;
    }

    public Solution() {
        desk = 0;
    }

    public Solution(int partialSolution) {
        this();
        addQueen(partialSolution);
    }

    long stack[] = new long[SIZE];
    int sp = 0;

    private boolean addQueen(int pos) {
        if (verts[pos])
            return false;

        if (diags1[currentLine - pos + (SIZE - 1)])
            return false;

        if (diags2[currentLine + pos])
            return false;

        stack[sp++] = desk;
        desk = Desk.setPos(desk, pos, currentLine);

        verts[pos] = diags1[currentLine - pos + (SIZE - 1)] = diags2[currentLine
                + pos] = true;

        currentLine++;
        return true;
    }

    private void removeLastQueen(int pos) {
        currentLine--;
        desk = stack[--sp];
        verts[pos] = diags1[currentLine - pos + (SIZE - 1)] = diags2[currentLine
                + pos] = false;
    }

    private boolean arrange() {
        if (currentLine == SIZE) {
            synchronized (known_positions) {
                if ((!CHECK_TRANSITIONS) || (!known_positions.contains(desk))) {
                    addKnownPosition(desk);
                    synchronized (solutions) {
                        solutions[solution_pointer++] = desk;
                        return true;
                    }
                }
            }
        } else {
            for (int x = 0; x < SIZE; x++) {
                if (addQueen(x)) {
                    if (arrange() && FIRST_ONLY)
                        return true;
                    removeLastQueen(x);
                }
            }
        }
        return false;
    }

    private static void addKnownPosition(long desk) {
        long clone = desk;
        long fliphor = Desk.cloneFlipHor(clone);
        long flipvert = Desk.cloneFlipVert(clone);

        for(int i=0;i<3;i++) {
            known_positions.add(clone);
            known_positions.add(fliphor);
            known_positions.add(flipvert);
    
            clone = Desk.cloneRotate(clone);
            fliphor = Desk.cloneFlipHor(clone);
            flipvert = Desk.cloneFlipVert(clone);
        }

        known_positions.add(clone);
        known_positions.add(fliphor);
        known_positions.add(flipvert);
    }

    static int SIZE;
    static int BITS_PER_CELL;
    static long FILL_MASK;
    static boolean FIRST_ONLY = false;
    static boolean CHECK_TRANSITIONS = true;

    static final Set<Long> known_positions = new HashSet<Long>(2000000);
    static final long solutions[] = new long[2000000];
    static int solution_pointer = 0;

    boolean verts[] = new boolean[SIZE];
    boolean diags1[] = new boolean[2 * SIZE];
    boolean diags2[] = new boolean[2 * SIZE];
    int currentLine = 0;
    long desk = 0;

}